<?php

namespace Drupal\Tests\vcp4dates\Kernel;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Form\FormState;
use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\user\Traits\UserCreationTrait;
use Drupal\views\Plugin\views\display\DisplayPluginBase;
use Drupal\views\ViewExecutable;
use Drupal\vcp4dates\Plugin\views\cache\DateFilter;
use Prophecy\PhpUnit\ProphecyTrait;

/**
 * Tests the DateFilter views cache plugin.
 *
 * @group vcp4dates
 */
class DateFilterTest extends KernelTestBase {

  use ProphecyTrait;
  use UserCreationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'user',
    'views',
    'vcp4dates',
  ];

  /**
   * The mocked time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface|\Prophecy\Prophecy\ObjectProphecy
   */
  protected $time;

  /**
   * The mocked view executable.
   *
   * @var \Drupal\views\ViewExecutable|\Prophecy\Prophecy\ObjectProphecy
   */
  protected $view;

  /**
   * The mocked display plugin.
   *
   * @var \Drupal\views\Plugin\views\display\DisplayPluginBase|\Prophecy\Prophecy\ObjectProphecy
   */
  protected $display;

  /**
   * The date filter plugin being tested.
   *
   * @var \Drupal\vcp4dates\Plugin\views\cache\DateFilter
   */
  protected $dateFilter;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installSchema('system', ['sequences']);

    // Mock the time service.
    $this->time = $this->prophesize(TimeInterface::class);
    $this->time->getRequestTime()->willReturn(1000000000);

    // Create the date filter plugin with mocked dependencies.
    $this->dateFilter = new DateFilter(
      ['date_filter' => 'test_date_filter'],
      'date_filter',
      ['provider' => 'vcp4dates'],
      $this->time->reveal()
    );

    // Mock the view and display.
    $this->view = $this->prophesize(ViewExecutable::class);
    $this->display = $this->prophesize(DisplayPluginBase::class);

    // Set up the view property on the plugin.
    $reflection = new \ReflectionClass($this->dateFilter);
    $property = $reflection->getProperty('view');
    $property->setAccessible(TRUE);
    $property->setValue($this->dateFilter, $this->view->reveal());
  }

  /**
   * Tests the defineOptions method.
   */
  public function testDefineOptions() {
    // Use reflection to access the protected method.
    $reflection = new \ReflectionClass($this->dateFilter);
    $method = $reflection->getMethod('defineOptions');
    $method->setAccessible(TRUE);
    $options = $method->invoke($this->dateFilter);

    $this->assertArrayHasKey('date_filter', $options);
    $this->assertEquals(NULL, $options['date_filter']['default']);
  }

  /**
   * Tests the summaryTitle method.
   */
  public function testSummaryTitle() {
    // Set up the options property on the plugin.
    $reflection = new \ReflectionClass($this->dateFilter);
    $property = $reflection->getProperty('options');
    $property->setAccessible(TRUE);
    $property->setValue($this->dateFilter, ['date_filter' => 'test_filter']);

    $this->assertEquals('test_filter', $this->dateFilter->summaryTitle());
  }

  /**
   * Tests the buildOptionsForm method.
   */
  public function testBuildOptionsForm() {
    // Set up the view to return the display.
    $this->view->getDisplay()->willReturn($this->display->reveal());

    // Set up the display to return filters.
    $filters = [
      'date_filter' => [
        'id' => 'date_filter',
        'plugin_id' => 'date',
        'operator' => '>',
      ],
      'other_filter' => [
        'id' => 'other_filter',
        'plugin_id' => 'string',
        'operator' => '=',
      ],
    ];
    $this->display->getOption('filters')->willReturn($filters);

    // Create a form and form state.
    $form = [];
    $form_state = new FormState();

    // Call the method.
    $this->dateFilter->buildOptionsForm($form, $form_state);

    // Check that the form has the expected elements.
    $this->assertArrayHasKey('date_filter', $form);
    $this->assertEquals('select', $form['date_filter']['#type']);
    $this->assertArrayHasKey('date_filter', $form['date_filter']['#options']);
    $this->assertArrayNotHasKey('other_filter', $form['date_filter']['#options']);
  }

  /**
   * Tests the cacheExpire method.
   */
  public function testCacheExpire() {
    // Set up the options property on the plugin.
    $reflection = new \ReflectionClass($this->dateFilter);
    $property = $reflection->getProperty('options');
    $property->setAccessible(TRUE);
    $property->setValue($this->dateFilter, ['date_filter' => 'test_date_filter']);

    // Mock the getLifespan method to return a specific value.
    $dateFilter = $this->getMockBuilder(DateFilter::class)
      ->setConstructorArgs([
        ['date_filter' => 'test_date_filter'],
        'date_filter',
        ['provider' => 'vcp4dates'],
        $this->time->reveal(),
      ])
      ->onlyMethods(['getLifespan'])
      ->getMock();
    $dateFilter->expects($this->once())
      ->method('getLifespan')
      ->willReturn(3600);

    // Set up the view property on the plugin.
    $reflection = new \ReflectionClass($dateFilter);
    $property = $reflection->getProperty('view');
    $property->setAccessible(TRUE);
    $property->setValue($dateFilter, $this->view->reveal());

    // Call the cacheExpire method using reflection.
    $method = $reflection->getMethod('cacheExpire');
    $method->setAccessible(TRUE);
    $result = $method->invoke($dateFilter, 'results');

    // The result should be the request time minus the lifespan.
    $this->assertEquals(1000000000 - 3600, $result);
  }

  /**
   * Tests the cacheSetMaxAge method.
   */
  public function testCacheSetMaxAge() {
    // Set up the options property on the plugin.
    $reflection = new \ReflectionClass($this->dateFilter);
    $property = $reflection->getProperty('options');
    $property->setAccessible(TRUE);
    $property->setValue($this->dateFilter, ['date_filter' => 'test_date_filter']);

    // Mock the getLifespan method to return a specific value.
    $dateFilter = $this->getMockBuilder(DateFilter::class)
      ->setConstructorArgs([
        ['date_filter' => 'test_date_filter'],
        'date_filter',
        ['provider' => 'vcp4dates'],
        $this->time->reveal(),
      ])
      ->onlyMethods(['getLifespan'])
      ->getMock();
    $dateFilter->expects($this->once())
      ->method('getLifespan')
      ->willReturn(3600);

    // Set up the view property on the plugin.
    $reflection = new \ReflectionClass($dateFilter);
    $property = $reflection->getProperty('view');
    $property->setAccessible(TRUE);
    $property->setValue($dateFilter, $this->view->reveal());

    // Call the cacheSetMaxAge method using reflection.
    $method = $reflection->getMethod('cacheSetMaxAge');
    $method->setAccessible(TRUE);
    $result = $method->invoke($dateFilter, 'results');

    // The result should be the lifespan.
    $this->assertEquals(3600, $result);
  }

  /**
   * Tests the getLifespan method.
   */
  public function testGetLifespan() {
    // Set up the options property on the plugin.
    $reflection = new \ReflectionClass($this->dateFilter);
    $property = $reflection->getProperty('options');
    $property->setAccessible(TRUE);
    $property->setValue($this->dateFilter, ['date_filter' => 'test_date_filter']);

    // Set up the view to return the display.
    $this->view->getDisplay()->willReturn($this->display->reveal());

    // Set up the display to return filters.
    $filters = [
      'test_date_filter' => [
        'id' => 'test_date_filter',
        'plugin_id' => 'date',
        'operator' => '>',
        'table' => 'node_field_data',
        'field' => 'created',
        'value' => [
          'value' => '2023-01-01',
        ],
      ],
    ];
    $this->display->getOption('filters')->willReturn($filters);

    // Mock the query and results.
    $query = $this->prophesize(\Drupal\Core\Database\Query\Select::class);
    $query_executable = $this->prophesize(\Drupal\views\Plugin\views\query\QueryPluginBase::class);
    $query_executable->query()->willReturn($query->reveal());
    $this->view->getQuery()->willReturn($query_executable->reveal());

    // Mock the addExpression method.
    $query->addExpression('node_field_data.created', 'dateFilterCachePluginValue')->willReturn($query->reveal());

    // Mock the execute and fetchAll methods.
    $statement = $this->prophesize(\Drupal\Core\Database\StatementInterface::class);
    $query->execute()->willReturn($statement->reveal());

    // Create a mock result object with a dateFilterCachePluginValue property.
    $result = new \stdClass();
    $result->dateFilterCachePluginValue = '2023-02-01';
    $statement->fetchAll()->willReturn([$result]);

    // Call the getLifespan method using reflection.
    $method = $reflection->getMethod('getLifespan');
    $method->setAccessible(TRUE);
    $result = $method->invoke($this->dateFilter);

    // The result should be the difference between the filter value and the result timestamp.
    // 2023-02-01 timestamp - 2023-01-01 timestamp
    $expected = strtotime('2023-02-01') - strtotime('2023-01-01');
    $this->assertEquals($expected, $result);
  }

}
